/* 

    TREELANG Compiler almost main (tree1)
    Called by gcc's toplev.c

    Copyright (C) 1986, 87, 89, 92-96, 1997, 1999, 2000, 2001 Free Software Foundation, Inc.

    This program is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
    Free Software Foundation; either version 2, or (at your option) any
    later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, 59 Temple Place - Suite 330,
    Boston, MA 02111-1307, USA.

    In other words, you are welcome to use, share and improve this program.
    You are forbidden to forbid anyone else to use, share and improve
    what you give them.   Help stamp out software-hoarding!  

    ---------------------------------------------------------------------------

    Written by Tim Josling 1999, 2000, 2001, based in part on other
    parts of the GCC compiler.

*/

#include "config.h"
#include "system.h"
#include "gansidecl.h"
#include "flags.h"
#include "output.h"
#include "toplev.h"

#include "ggc.h"
#include "tree.h"

#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <stdarg.h>
#include <string.h>
#include <stdio.h>

#include "treelang.h"

/* Each front end provides its own.  */
static void tree_init (void);
static void tree_finish (void);
static void tree_init_options (void);
static int tree_decode_option (int, char**);
static void tree_post_options (void);

/* Linked list of symbols - all must be unique in treelang.  */

struct production* symbol_table = NULL;

/* Hooks for back end and toplev.c.  */

struct lang_hooks lang_hooks = 
{tree_init,
 tree_finish, 
 tree_init_options,
 tree_decode_option,
 tree_post_options};

/* Language for usage for messages.  */

const char* const language_string = "TREELANG - sample gcc front end for GCC weekly snapshot 20010423 of ";

/* Local prototypes.  */

void version (void);

/* Fake parse routine in this module - name dictated by toplev.c.  */
int yyparse (void);

/* Real parse module in treeprs.c.  */
int tree_parse (void);

/* GC routine for symbol table.  */
static void symbol_table_ggc (void* m);

/* Global variables.  */

extern int errorcount; 

extern struct cbl_tree_struct_parse_tree_top* parse_tree_top;

/* 
   Options. 
*/

/* Trace the parser.  */

extern int yydebug;
unsigned int option_parser_trace = 0;

/* Trace the lexical analysis.  */

unsigned int option_lexer_trace = 0;

/* Warning levels.  */

/* Local variables.  */

unsigned char* in_fname = NULL;	/* Input file name.  */

/* This is 1 if we have output the version string.  */

static int version_done = 0;

/* Variable nesting level.  */

static unsigned int current_nesting_level = 0;

/* Process one switch - called by toplev.c.  */

int
tree_decode_option (num_options_left, first_option_left)
     int num_options_left ATTRIBUTE_UNUSED; 
     char** first_option_left;
{
  
  /*
    Process options - bear in mind I may get options that are really
    meant for someone else (eg the main compiler) so I have to be very
    permissive. 
    
  */
  
  if (first_option_left[0][0] != '-')
    return 0; 
  
  switch (first_option_left[0][1]) 
    {
    case '-':
      if (!strcmp (first_option_left[0],"--help"))
        {
          if (!version_done)
            {
              fputs (language_string, stdout);
              fputs (version_string, stdout);
              fputs ("\n", stdout);
              version_done = 1;
            }
          fprintf (stdout, "Usage: tree1 [switches] -o output input\n");
          return 1;
        }
    case 'v':
      if (!strcmp (first_option_left[0],"-v"))
        {
          if (!version_done)
            {
              fputs (language_string, stdout);
              fputs (version_string, stdout);
              fputs ("\n", stdout);
              version_done = 1;
            }
          return 1;
        }
    case 'f':
      if (!strcmp (first_option_left[0],"-fparser-trace"))
        {
          option_parser_trace = 1;
          yydebug = 1;
          return 1;
        }
      if (!strcmp (first_option_left[0],"-flexer-trace"))
        {
          option_lexer_trace = 1;
          return 1;
        }
      return 0;

    case 'w':
      if (!strcmp (first_option_left[0],"-w"))
        {
          /* Tolerate this option but ignore it - we always put out
             all warnings.  */
          return 1;
        }
      return 0;

    case 'W':
      if (!strcmp (first_option_left[0],"-Wall"))
        {
          return 1;
        }
      return 0;

    default:
      return 0;
    }

  return 0;

}

/* Initialize options.  */

void 
tree_init_options (void)
{
}

/* Wrap up options.  */

void 
tree_post_options (void)
{
}

/* Performs whatever initialization steps are needed by the
   language-dependent front end.  */

void
tree_init (void)
{
  /* Define my garbage collection routines.  */
  ggc_add_root (&symbol_table, 1, 
                /* Unused size.  */ sizeof (void*), symbol_table_ggc);
  /* Note: only storage that has to be kept across functions needs to
     be protected from GC.  */
}

/* Performs whatever wrapup steps are needed by the language-dependent
   front end. */

void
tree_finish (void)
{
}

/* Language dependent parser setup.  */

const char*
init_parse (const char* filename)
{
  /* This error will not happen from gcc as it will always create a
     fake input file.  */
  if (!filename || (filename[0] == ' ') || (!filename[0])) 
    {
      if (!version_done)
        {
          fprintf (stderr, "No input file specified, try --help for help\n");
          exit (1);
        }

      in_fname = NULL;
      return NULL;
    }
  yyin = fopen (filename, "r");
  if (!yyin)
    {
      fprintf (stderr, "Unable to open input file %s\n", filename);
      exit (1);
    }
  return (char*) (in_fname = (unsigned char*)filename);
}

/* Language dependent wrapup.  */

void 
finish_parse (void)
{
  fclose (yyin);
}

/* Return a short string identifying this language to the debugger.  */

const char*
lang_identify (void)
{ 
  return "treelang";
}

/* Print language-specific statistics.  */

void
print_lang_statistics (void)
{
}

/* Scan the symbol table* M, marking storage used.  */

static void
symbol_table_ggc (void* m)
{
  struct production* pp;
  pp = * (struct production**)m;
  /* Actually it is a pointer to a pointer, to allow reallocation and
     relinking.  */
  mark_production_used (pp);
}

/* Mark a production PP as used so it wont be garbage collected.  */

void
mark_production_used (struct production* pp)
{
  int sub_ix;
 loop:
  if (!pp)
    return;
  ggc_mark (pp);
  
  if (pp->category == token_category)
    {
      mark_token_used ((struct token*)pp);
      return;
    }
  if (pp->category != production_category)
    abort ();
  mark_token_used (pp->main_token);
  for (sub_ix = 0; sub_ix < SUB_COUNT; sub_ix++)
    mark_production_used (pp->sub[sub_ix]);
  /* The macro tests for NULL so I don't need to.  */
  ggc_mark_tree (pp->code);
  pp = pp->next;
  goto loop;
}

/* Mark a token TT as used so it wont be garbage collected.  */

void
mark_token_used (struct token* tt)
{
  if (!tt) 
    return;
  ggc_mark (tt);
  if (tt->chars)
    ggc_mark (tt->chars);
}

/* Set debug (-y) option.  */

void
set_yydebug (int value ATTRIBUTE_UNUSED)
{
  option_parser_trace = 1;
  option_lexer_trace = 1;
}

/* Allocate SIZE bytes and clear them.  */

void*
my_malloc (size_t size)
{
  void* mem;
  mem = ggc_alloc (size);
  if (!mem)
    {
      fprintf (stderr, "\nOut of memory\n");
      abort ();
    }
  memset (mem,0,size);
  return mem;
}

/* Look up a name in PROD->SYMBOL_TABLE_NAME in the symbol table;
   return the symbol table entry from the symbol table if found there,
   else 0.  */

struct production*
lookup_tree_name (struct production* prod)
{
  struct production* this;
  struct token* this_tok;
  struct token* tok;
  tok = SYMBOL_TABLE_NAME (prod);
  for (this = symbol_table; this; this = this->next)
    {
      this_tok = this->main_token;
      if (tok->length != this_tok->length) 
        continue;
      if (memcmp (tok->chars, this_tok->chars, this_tok->length))
        continue;
      if (option_parser_trace)
        fprintf (stderr, "Found symbol %s (%i:%i) as %i \n", tok->chars, 
                tok->lineno, tok->charno, NUMERIC_TYPE (this));
      return this;
    }
  if (option_parser_trace)
    fprintf (stderr, "Not found symbol %s (%i:%i) as %i \n", tok->chars, 
            tok->lineno, tok->charno, tok->type);
  return NULL;
}

/* Insert name PROD into the symbol table.  Return 1 if duplicate, 0 if OK.  */

int
insert_tree_name (struct production* prod)
{
  struct token* tok;
  tok = SYMBOL_TABLE_NAME (prod);
  if (lookup_tree_name (prod))
    {
      fprintf (stderr, "%s:%i:%i duplicate name %s\n", in_fname, tok->lineno, tok->charno, tok->chars);
      errorcount++;
      return 1;
    }
  prod->next = symbol_table;
  NESTING_LEVEL (prod) = current_nesting_level;
  symbol_table = prod;
  return 0;
}

void
push_var_level (void)
{
  current_nesting_level++;
}

/* Pop variable nesting level and remove those variable from symbol table.  */

void
pop_var_level (void)
{
  current_nesting_level--;

  while (symbol_table)
    {
      if (NESTING_LEVEL (symbol_table) > current_nesting_level)
        symbol_table = symbol_table->next;
      else
        break;
    }
}


/* Create a struct productions of type TYPE, main token MAIN_TOK.  */

struct production* 
make_production (int type, struct token* main_tok)
{
  struct production* prod;
  prod = my_malloc (sizeof (struct production));
  prod->category = production_category;
  prod->type = type;
  prod->main_token = main_tok;
  return prod;
}
